package rule

import (
	"code.fnuoos.com/go_rely_warehouse/zyos_go_order_relate_rule.git/db"
	"code.fnuoos.com/go_rely_warehouse/zyos_go_order_relate_rule.git/db/model"
	"code.fnuoos.com/go_rely_warehouse/zyos_go_order_relate_rule.git/enum"
	"code.fnuoos.com/go_rely_warehouse/zyos_go_order_relate_rule.git/md"
	"code.fnuoos.com/go_rely_warehouse/zyos_go_order_relate_rule.git/svc"
	zhios_order_relate_utils "code.fnuoos.com/go_rely_warehouse/zyos_go_order_relate_rule.git/utils"
	"code.fnuoos.com/go_rely_warehouse/zyos_go_order_relate_rule.git/utils/cache"
	zhios_order_relate_logx "code.fnuoos.com/go_rely_warehouse/zyos_go_order_relate_rule.git/utils/logx"
	"errors"
	"fmt"
	"github.com/shopspring/decimal"
	"strconv"
	"time"
	"xorm.io/xorm"
)

func InitForConsumeIntegral(redisAddr string) (err error) {
	if redisAddr != "" {
		cache.NewRedis(redisAddr)
	}
	_, err = cache.SelectDb(md.RedisDataBase)
	return
}

const PessimismLockKeyForConsumeIntegral = "daily_settlement_consume_integral_lock_key:%s"
const PessimismLockValueForConsumeIntegral = "running"

// DailySettlementBlockConsumeIntegral 每日结算“消费积分”
func DailySettlementBlockConsumeIntegral(engine *xorm.Engine, mid string, isTask bool) (err error) {
	session := engine.NewSession()
	defer func() {
		session.Close()
		if err := recover(); err != nil {

			_ = zhios_order_relate_logx.Error(err)
		}
	}()
	session.Begin()
	now := time.Now()
	today := now.Format("2006-01-02")

	//1、查找 `subsidy_base` 基础设置
	subsidyBase, err := db.SubsidyBaseGetOneByParams(session, map[string]interface{}{
		"key":   "is_open",
		"value": 1,
	})

	fmt.Println(">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>", now.Hour())
	settleTime, err := time.ParseInLocation("2006-01-02 15:04:05", today+" "+subsidyBase.SettleTime+":00", time.Local)
	if err != nil {
		session.Rollback()
		return
	}
	if isTask && (now.Hour() > 8 || now.Before(settleTime)) {
		session.Rollback()
		//TODO::结算时间 ~ 凌晨 8 点运行
		return errors.New("非运行时间")
	}
	if isTask && zhios_order_relate_utils.InArr(now.Weekday().String(), []string{"Sunday", "Saturday"}) && subsidyBase.HolidaySettleOpen != 1 {
		session.Rollback()
		return errors.New("周末不运行")
	}
	if err != nil {
		_ = session.Rollback()
		return err
	}
	if isTask && subsidyBase.SettlementDate == today {
		_ = session.Rollback()
		return errors.New("今日“消费计费”已结算")
	}
	key := fmt.Sprintf(PessimismLockKeyForConsumeIntegral, mid)
	//TODO::增加“悲观锁”防止串行
	getString, _ := cache.GetString(key)
	//if err != nil {
	//	return err
	//}
	if getString == PessimismLockValueForConsumeIntegral {
		fmt.Println("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!", "上一次结算未执行完")
		return errors.New("上一次结算未执行完")
	}
	cache.SetEx(key, PessimismLockValueForConsumeIntegral, 3600*8) //8小时

	var consumeValueTotal, experienceValueTotal = decimal.NewFromInt(0), decimal.NewFromInt(0)
	//1、统计当前拥有多少份消费补贴
	startAt := time.Now().Add(time.Duration(subsidyBase.ConsumptionDay) * -24 * time.Hour).Format("2006-01-02") //起始时间
	var subsidyWithUserForConsumeList []model.SubsidyWithUser
	hasConsumeTotal, err := session.Where("date <= ?", startAt).And("kind = ? and is_can_subsidy=?", 1, 1).FindAndCount(&subsidyWithUserForConsumeList)
	if err != nil {
		_ = session.Rollback()

		return
	}
	if hasConsumeTotal > 0 {
		consumeValue, err := calcNowEverydayConsumeIntegral(subsidyBase, hasConsumeTotal)
		if err != nil {
			_ = session.Rollback()
			return err
		}

		for _, v := range subsidyWithUserForConsumeList {
			consumeValueTotal = consumeValueTotal.Add(decimal.NewFromFloat(consumeValue))
			err1 := DealUserConsumeIntegral(session, &v, consumeValue, mid)
			if err1 != nil {
				_ = session.Rollback()
				return err1
			}
		}
	}
	//2、统计当前拥有多少份体验补贴
	var subsidyWithUserForExperienceList []model.SubsidyWithUser
	hasExperienceTotal, err := session.Where("date <= ?", startAt).And("kind = ? and is_can_subsidy=?", 2, 1).FindAndCount(&subsidyWithUserForExperienceList)
	if err != nil {
		_ = session.Rollback()
		return
	}
	if hasExperienceTotal > 0 {
		experienceValue, err := calcNowEverydayExperienceIntegral(subsidyBase, hasExperienceTotal)
		if err != nil {
			_ = session.Rollback()
			return err
		}
		for _, v := range subsidyWithUserForConsumeList {
			experienceValueTotal = experienceValueTotal.Add(decimal.NewFromFloat(experienceValue))
			err1 := DealUserExperienceIntegral(session, &v, experienceValue, mid)
			if err1 != nil {
				_ = session.Rollback()
				return err1
			}
		}
	}
	//3、修改 subsidy_base 中 consumption_money、experience_money
	consumeTotal, _ := consumeValueTotal.Float64()
	experienceTotal, _ := experienceValueTotal.Float64()
	subsidyBase.ConsumptionTotalMoney = zhios_order_relate_utils.Float64ToStr(zhios_order_relate_utils.StrToFloat64(subsidyBase.ConsumptionTotalMoney) - consumeTotal)
	subsidyBase.ExperienceTotalMoney = zhios_order_relate_utils.Float64ToStr(zhios_order_relate_utils.StrToFloat64(subsidyBase.ExperienceTotalMoney) - experienceTotal)
	subsidyBase.SettlementDate = today
	updateAffected, err := db.SubsidyBaseUpdate(session, subsidyBase.Id, subsidyBase, "consumption_total_money", "experience_total_money", "settlement_date")
	if err != nil {
		_ = session.Rollback()
		return
	}
	if updateAffected <= 0 {
		_ = session.Rollback()
		return errors.New("修改 subsidy_base 失败")
	}

	//4、提交事务
	err = session.Commit()
	if err != nil {
		_ = session.Rollback()
		return errors.New("事务提交失败")
	}

	cache.Del(key)
	fmt.Println(">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>消费积分结束<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<")
	return nil
}

//计算当前消费积分价值（公式 "((资金池的金额 * 浮动比例) / 消费补贴总数) + 基础补贴"）
func calcNowEverydayConsumeIntegral(subsidyBase *model.SubsidyBase, hasConsumeTotal int64) (value float64, err error) {
	consumptionMoney, _ := decimal.NewFromString(subsidyBase.ConsumptionTotalMoney)
	baseSubsidyMoney, _ := decimal.NewFromString(subsidyBase.BaseSubsidyMoney)
	floatSubsidyOpen := decimal.NewFromInt(int64(subsidyBase.FloatSubsidyOpen) / 10000)
	consumeTotal := decimal.NewFromInt(hasConsumeTotal)
	value, _ = consumptionMoney.Mul(floatSubsidyOpen).Div(consumeTotal).Add(baseSubsidyMoney).Float64()
	return
}

//计算当前体验积分价值
func calcNowEverydayExperienceIntegral(subsidyBase *model.SubsidyBase, hasExperienceTotal int64) (value float64, err error) {
	//2、通过公式计算 "((资金池的金额 * 浮动比例) / 体验补贴总数) + 基础补贴"
	experienceMoney, _ := decimal.NewFromString(subsidyBase.ExperienceTotalMoney)
	baseSubsidyMoney, _ := decimal.NewFromString(subsidyBase.BaseSubsidyMoney)
	floatSubsidyOpen := decimal.NewFromInt(int64(subsidyBase.FloatSubsidyOpen) / 10000)
	consumeTotal := decimal.NewFromInt(hasExperienceTotal)
	value, _ = experienceMoney.Mul(floatSubsidyOpen).Div(consumeTotal).Add(baseSubsidyMoney).Float64()
	return
}

// DealUserConsumeIntegral 处理给用户发放消费补贴奖励
func DealUserConsumeIntegral(session *xorm.Session, subsidyWithUser *model.SubsidyWithUser, consumeIntegralValue float64, mid string) (err error) {
	now := time.Now()
	//1、扣除 subsidy_with_user 的 balance_amount、state
	if subsidyWithUser.Amount == subsidyWithUser.BalanceAmount {
		subsidyWithUser.State = 2 //补贴中
	}
	balanceAmount, _ := decimal.NewFromString(subsidyWithUser.BalanceAmount)
	afterAmount, _ := balanceAmount.Sub(decimal.NewFromFloat(consumeIntegralValue)).Float64()
	if afterAmount <= 0 {
		afterAmount = 0
		consumeIntegralValue = zhios_order_relate_utils.StrToFloat64(subsidyWithUser.BalanceAmount)
		subsidyWithUser.State = 3 //补贴完
	}
	subsidyWithUser.BalanceAmount = zhios_order_relate_utils.Float64ToStr(afterAmount)
	updateAffected, err := db.SubsidyWithUserUpdate(session, subsidyWithUser.Id, subsidyWithUser, "balance_amount", "state")
	if err != nil {
		return err
	}
	if updateAffected <= 0 {
		return errors.New("修改 subsidy_with_user 记录失败")
	}

	//2、新增 subsidy_with_user_flow 记录
	var subsidyWitUserFlow = &model.SubsidyWithUserFlow{
		RecordsId:     subsidyWithUser.Id,
		Amount:        zhios_order_relate_utils.Float64ToStr(consumeIntegralValue),
		BalanceAmount: zhios_order_relate_utils.Float64ToStr(afterAmount),
		CreateAt:      now.Format("2006-01-02 15:04:05"),
		UpdateAt:      now.Format("2006-01-02 15:04:05"),
		Uid:           subsidyWithUser.Uid,
		Kind:          subsidyWithUser.Kind,
		Month:         zhios_order_relate_utils.StrToInt(now.Format("200601")),
	}
	insertAffected, err := db.SubsidyWithUserFlowInsert(session, subsidyWitUserFlow)
	if err != nil {
		return err
	}
	if insertAffected <= 0 {
		return errors.New("新增 subsidy_with_user_flow 记录失败")
	}
	//3、加入每月统计
	if consumeIntegralValue > 0 {
		var subsidyWitUserMonth = &model.SubsidyWithUserMonth{
			Uid:           subsidyWithUser.Uid,
			ConsumeAmount: zhios_order_relate_utils.Float64ToStr(consumeIntegralValue),
			Date:          now.Format("2006-01"),
		}
		_, err1 := db.SubsidyWithMonthInsert(session, subsidyWitUserMonth)
		if err1 != nil {
			return err
		}
	}
	//4、给用户添加余额
	orderType := enum.FinUserFlowOrderActionString(md.ConsumeSubsidyOrdActionForFinUserFlow)
	var dealUserAmount = md.DealUserAmount{
		Kind:      "add",
		Mid:       mid,
		Title:     md.ConsumeSubsidyTitleForFinUserFlow,
		OrderType: orderType,
		OrdAction: md.ConsumeSubsidyOrdActionForFinUserFlow,
		OrdId:     strconv.Itoa(subsidyWithUser.Id),
		Uid:       subsidyWithUser.Uid,
		Amount:    consumeIntegralValue,
	}
	err = svc.DealUserAmount(session, dealUserAmount)
	if err != nil {
		return
	}
	return nil
}

// DealUserExperienceIntegral 处理给用户发放体验补贴奖励
func DealUserExperienceIntegral(session *xorm.Session, subsidyWithUser *model.SubsidyWithUser, experienceIntegralValue float64, mid string) (err error) {
	now := time.Now()
	//1、扣除 subsidy_with_user 的 balance_amount、state
	if subsidyWithUser.Amount == subsidyWithUser.BalanceAmount {
		subsidyWithUser.State = 2 //补贴中
	}
	balanceAmount, _ := decimal.NewFromString(subsidyWithUser.BalanceAmount)
	afterAmount, _ := balanceAmount.Sub(decimal.NewFromFloat(experienceIntegralValue)).Float64()
	if afterAmount <= 0 {
		afterAmount = 0
		experienceIntegralValue = zhios_order_relate_utils.StrToFloat64(subsidyWithUser.BalanceAmount)
		subsidyWithUser.State = 3 //补贴完
	}
	subsidyWithUser.BalanceAmount = zhios_order_relate_utils.Float64ToStr(afterAmount)
	updateAffected, err := db.SubsidyWithUserUpdate(session, subsidyWithUser.Id, subsidyWithUser, "balance_amount", "state")
	if err != nil {
		return err
	}
	if updateAffected <= 0 {
		return errors.New("修改 subsidy_with_user 记录失败")
	}

	//2、新增 subsidy_with_user_flow 记录
	var subsidyWitUserFlow = &model.SubsidyWithUserFlow{
		RecordsId:     subsidyWithUser.Id,
		Uid:           subsidyWithUser.Uid,
		Kind:          subsidyWithUser.Kind,
		Amount:        zhios_order_relate_utils.Float64ToStr(experienceIntegralValue),
		BalanceAmount: zhios_order_relate_utils.Float64ToStr(afterAmount),
		CreateAt:      now.Format("2006-01-02 15:04:05"),
		Month:         zhios_order_relate_utils.StrToInt(now.Format("200601")),
		UpdateAt:      now.Format("2006-01-02 15:04:05"),
	}
	insertAffected, err := db.SubsidyWithUserFlowInsert(session, subsidyWitUserFlow)
	if err != nil {
		return err
	}
	if insertAffected <= 0 {
		return errors.New("新增 subsidy_with_user_flow 记录失败")
	}
	//3、加入每月统计
	if experienceIntegralValue > 0 {
		var subsidyWitUserMonth = &model.SubsidyWithUserMonth{
			Uid:              subsidyWithUser.Uid,
			ExperienceAmount: zhios_order_relate_utils.Float64ToStr(experienceIntegralValue),
			Date:             now.Format("2006-01"),
		}
		_, err1 := db.SubsidyWithMonthInsert(session, subsidyWitUserMonth)
		if err1 != nil {
			return err
		}
	}
	//4、给用户添加余额
	orderType := enum.FinUserFlowOrderActionString(md.ExperienceSubsidyOrdActionForFinUserFlow)
	var dealUserAmount = md.DealUserAmount{
		Kind:      "add",
		Mid:       mid,
		Title:     md.ExperienceSubsidyTitleForFinUserFlow,
		OrderType: orderType,
		OrdAction: md.ExperienceSubsidyOrdActionForFinUserFlow,
		OrdId:     strconv.Itoa(subsidyWithUser.Id),
		Uid:       subsidyWithUser.Uid,
		Amount:    experienceIntegralValue,
	}
	err = svc.DealUserAmount(session, dealUserAmount)
	if err != nil {
		return
	}
	return nil

}
